from logging import ERROR, INFO
from typing import Optional
import click
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.api.utils._roles import setup_role
from taf.messages import git_commit_message
from taf.models.types import RolesIterator
from taf.models.types import RolesKeysData
from taf.api.utils._git import commit_and_push
from taf.models.converter import from_dict

from pathlib import Path
from taf.api.roles import (
    create_delegations,
    _initialize_roles_and_keystore,
)
from taf.api.targets import register_target_files

from taf.auth_repo import AuthenticationRepository
from taf.exceptions import TAFError
from taf.keys import load_sorted_keys_of_new_roles
from tuf.repository_tool import create_new_repository
from taf.log import taf_logger


@log_on_start(
    INFO, "Creating a new authentication repository {path:s}", logger=taf_logger
)
@log_on_end(INFO, "Finished creating a new repository", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while creating a new repository: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def create_repository(
    path: str,
    keystore: Optional[str] = None,
    roles_key_infos: Optional[str] = None,
    commit: Optional[bool] = False,
    test: Optional[bool] = False,
) -> None:
    """
    Create a new authentication repository. Generate initial metadata files.
    If target files already exist, add corresponding targets information to
    targets metadata files.

    Arguments:
        path: Authentication repository's location.
        keystore: Location of the keystore files.
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        test (optional): Specifies if the created repository is a test authentication repository.

    Side Effects:
        Creates a new authentication repository (initializes a new git repository and generates tuf metadata)

    Returns:
        None
    """
    auth_repo = AuthenticationRepository(path=path)

    if not _check_if_can_create_repository(auth_repo):
        return

    roles_key_infos_dict, keystore = _initialize_roles_and_keystore(
        roles_key_infos, keystore
    )

    keystore_path = Path(keystore)
    if not keystore_path.is_dir():
        keystore_path.mkdir(parents=False)

    roles_keys_data = from_dict(roles_key_infos_dict, RolesKeysData)
    repository = create_new_repository(
        str(auth_repo.path), repository_name=auth_repo.name
    )
    signing_keys, verification_keys = load_sorted_keys_of_new_roles(
        auth_repo=auth_repo,
        roles=roles_keys_data.roles,
        yubikeys_data=roles_keys_data.yubikeys,
        keystore=keystore,
    )
    if signing_keys is None:
        return
    # set threshold and register keys of main roles
    # we cannot do the same for the delegated roles until delegations are created
    for role in RolesIterator(roles_keys_data.roles, include_delegations=False):
        setup_role(
            role,
            repository,
            verification_keys[role.name],
            signing_keys.get(role.name),
        )

    create_delegations(
        roles_keys_data.roles.targets, repository, verification_keys, signing_keys
    )

    # if the repository is a test repository, add a target file called test-auth-repo
    if test:
        test_auth_file = (
            Path(auth_repo.path, auth_repo.targets_path) / auth_repo.TEST_REPO_FLAG_FILE
        )
        test_auth_file.touch()

    # register and sign target files (if any)
    auth_repo._tuf_repository = repository
    updated = register_target_files(
        path,
        keystore,
        roles_key_infos,
        commit=False,
        taf_repo=auth_repo,
        write=True,
        no_commit_warning=True,
    )
    if not updated:
        repository.writeall()

    if commit:
        auth_repo.init_repo()
        commit_msg = git_commit_message("create-repo")
        commit_and_push(auth_repo, push=False, commit_msg=commit_msg)
    else:
        print("\nPlease commit manually.\n")


def _check_if_can_create_repository(auth_repo: AuthenticationRepository) -> bool:
    """
    Check if a new authentication repository can be created at the specified location.
    A repository can be created if there is not directory at the repository's location
    or if it does exists, is not the root of a git repository.

    Arguments:
        auth_repo: Authentication repository.

    Side Effects:
        None

    Returns:
        True if a new authentication repository can be created, False otherwise.
    """
    path = Path(auth_repo.path)
    if path.is_dir():
        # check if there is non-empty metadata directory
        if auth_repo.metadata_path.is_dir() and any(auth_repo.metadata_path.iterdir()):
            if auth_repo.is_git_repository:
                print(
                    f'"{path}" is a git repository containing the metadata directory. Generating new metadata files could make the repository invalid. Aborting.'
                )
                return False
            if not click.confirm(
                f'Metadata directory found inside "{path}". Recreate metadata files?'
            ):
                return False
    return True



from logging import DEBUG, ERROR, INFO
from typing import Dict, Optional
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.exceptions import TAFError
from taf.keys import load_signing_keys
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.repository_tool import Repository
from taf.log import taf_logger


@log_on_end(INFO, "Updated snapshot and timestamp", logger=taf_logger)
@log_on_error(
    ERROR,
    "Could not update snapshot and timestamp: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def update_snapshot_and_timestamp(
    taf_repo: Repository,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    write_all: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
) -> None:
    """
    Sign snapshot and timestamp metadata files.

    Arguments:
        taf_repo: Authentication repository.
        keystore: Keystore directory's path.
        scheme (optional): Signature scheme.
        write_all (optional): If True, writes authentication repository's
            changes to disk.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        push (optional): Flag specifying whether to push to remote

    Side Effects:
        Updates metadata files, saves changes to disk if write_all is True

    Returns:
        None
    """
    loaded_yubikeys: Dict = {}

    for role in ("snapshot", "timestamp"):
        keystore_keys, yubikeys = load_signing_keys(
            taf_repo,
            role,
            loaded_yubikeys,
            keystore,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
        )
        if len(yubikeys):
            update_method = taf_repo.roles_yubikeys_update_method(role)
            update_method(yubikeys, write=False)
        if len(keystore_keys):
            update_method = taf_repo.roles_keystore_update_method(role)
            update_method(keystore_keys, write=False)

    if write_all:
        taf_repo.writeall()


@log_on_start(DEBUG, "Updating target metadata", logger=taf_logger)
@log_on_end(DEBUG, "Updated target metadata", logger=taf_logger)
@log_on_error(
    ERROR,
    "Could not update target metadata: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def update_target_metadata(
    taf_repo: Repository,
    added_targets_data: Dict,
    removed_targets_data: Dict,
    keystore: str,
    write: Optional[bool] = False,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
) -> bool:
    """Given dictionaries containing targets that should be added and targets that should
    be removed, update and sign target metadata files and, if write is True, also
    sign snapshot and timestamp.

    Arguments:
        taf_repo: Authentication repository.
        added_targets_data(dict): Dictionary containing targets data that should be added.
        removed_targets_data(dict): Dictionary containing targets data that should be removed.
        keystore: Keystore directory's path.
        write (optional): If True, updates snapshot and timestamp and write changes to disk.
        scheme (optional): Signature scheme.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.

    Side Effects:
        Updates metadata files, saves changes to disk if write_all is True

    Returns:
        True if there were targets that were updated, False otherwise
    """
    added_targets_data = {} if added_targets_data is None else added_targets_data
    removed_targets_data = {} if removed_targets_data is None else removed_targets_data

    roles_targets = taf_repo.roles_targets_for_filenames(
        list(added_targets_data.keys()) + list(removed_targets_data.keys())
    )

    if not roles_targets:
        taf_logger.info("No target files to sign")
        return False

    # update targets
    loaded_yubikeys: Dict = {}
    for role, target_paths in roles_targets.items():
        keystore_keys, yubikeys = load_signing_keys(
            taf_repo,
            role,
            loaded_yubikeys,
            keystore,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
        )
        targets_data = dict(
            added_targets_data={
                path: val
                for path, val in added_targets_data.items()
                if path in target_paths
            },
            removed_targets_data={
                path: val
                for path, val in removed_targets_data.items()
                if path in target_paths
            },
        )

        if len(yubikeys):
            taf_repo.update_targets_yubikeys(yubikeys, write=False, **targets_data)
        if len(keystore_keys):
            taf_repo.update_targets_keystores(
                keystore_keys, write=False, **targets_data
            )

    if write:
        update_snapshot_and_timestamp(
            taf_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
        )
    return True



from datetime import datetime
from logging import INFO, ERROR
from typing import Dict, List, Optional, Tuple
from logdecorator import log_on_end, log_on_error
from taf.api.utils._git import check_if_clean, commit_and_push
from taf.exceptions import TAFError
from taf.git import GitRepository
from taf.keys import load_signing_keys
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.messages import git_commit_message
from taf.repository_tool import Repository, is_delegated_role
from taf.log import taf_logger


@log_on_error(
    ERROR,
    "An error occurred while checking expiration dates: {e!r}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=False,
)
def check_expiration_dates(
    path: str,
    interval: Optional[int] = 30,
    start_date: Optional[datetime] = None,
    excluded_roles: Optional[List[str]] = None,
    print_output: bool = True,
) -> Tuple[Dict, Dict]:
    """
    Check if any metadata files (roles) are expired or will expire in the next <interval> days.
    Prints a list of expired roles.

    Arguments:
        path: Authentication repository's location.
        interval: Number of days ahead to check for expiration.
        start_date: Date from which to start checking for expiration.
        excluded_roles: List of roles to exclude from the check.

    Side Effects:
        Prints lists of roles that expired or are about to expire.

    Returns:
        None
    """
    taf_repo = Repository(path)

    if start_date is None:
        start_date = datetime.now()

    expired_dict, will_expire_dict = taf_repo.check_roles_expiration_dates(
        interval, start_date, excluded_roles
    )

    if print_output:
        print_expiration_dates(
            expired_dict, will_expire_dict, start_date=start_date, interval=interval
        )

    return expired_dict, will_expire_dict


def print_expiration_dates(
    expired: Dict, will_expire: Dict, start_date: datetime, interval: Optional[int] = 30
) -> None:
    if expired or will_expire:
        now = datetime.now()
        print(
            f"Given a {interval} day interval from ({start_date.strftime('%Y-%m-%d')}):"
        )
        for role, expiry_date in expired.items():
            delta = now - expiry_date
            print(
                f"{role} expired {delta.days} days ago (on {expiry_date.strftime('%Y-%m-%d')})"
            )
        for role, expiry_date in will_expire.items():
            delta = expiry_date - now
            print(
                f"{role} will expire in {delta.days} days (on {expiry_date.strftime('%Y-%m-%d')})"
            )
    else:
        print(f"No roles will expire within the given {interval} day interval")


@check_if_clean
def update_metadata_expiration_date(
    path: str,
    roles: List[str],
    interval: int,
    keystore: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    start_date: Optional[datetime] = None,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
) -> None:
    """
    Update expiration dates of the specified roles and all other roles that need
    to be signed in order to guarantee validity of the repository e.g. snapshot
    and timestamp need to be signed after a targets role is updated.

    Arguments:
        path: Authentication repository's location.
        roles: A list of roles whose expiration dates should be updated.
        interval: Number of days added to the start date in order to calculate the
            expiration date.
        keystore (optional): Keystore directory's path
        scheme (optional): Signature scheme.
        start_date (optional): Date to which expiration interval is added.
            Set to today if not specified.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        push (optional): Flag specifying whether to push to remote

    Side Effects:
        Updates metadata files, saves changes to disk and commits changes
        unless no_commit is set to True.

    Returns:
        None
    """
    if start_date is None:
        start_date = datetime.now()

    taf_repo = Repository(path)
    loaded_yubikeys: Dict = {}
    roles_to_update = []

    if "root" in roles:
        roles_to_update.append("root")
    if "targets" in roles:
        roles_to_update.append("targets")
    for role in roles:
        if is_delegated_role(role):
            roles_to_update.append(role)

    if len(roles_to_update) or "snapshot" in roles:
        roles_to_update.append("snapshot")
    roles_to_update.append("timestamp")

    for role in roles_to_update:
        _update_expiration_date_of_role(
            taf_repo,
            role,
            loaded_yubikeys,
            keystore,
            start_date,
            interval,
            scheme,
            prompt_for_keys,
        )

    if not commit:
        print("\nPlease commit manually.\n")
    else:
        auth_repo = GitRepository(path=path)
        commit_msg = git_commit_message(
            "update-expiration-dates", roles=",".join(roles)
        )
        commit_and_push(auth_repo, commit_msg=commit_msg, push=push)


@log_on_end(INFO, "Updated expiration date of {role:s}", logger=taf_logger)
@log_on_error(
    ERROR,
    "Error: could not update expiration date: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def _update_expiration_date_of_role(
    auth_repo: Repository,
    role: str,
    loaded_yubikeys: Dict,
    keystore: str,
    start_date: datetime,
    interval: int,
    scheme: str,
    prompt_for_keys: bool,
) -> None:
    keys, yubikeys = load_signing_keys(
        auth_repo,
        role,
        loaded_yubikeys=loaded_yubikeys,
        keystore=keystore,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
    )
    # sign with keystore
    if len(keys):
        auth_repo.update_role_keystores(
            role, keys, start_date=start_date, interval=interval
        )
    if len(yubikeys):  # sign with yubikey
        auth_repo.update_role_yubikeys(
            role, yubikeys, start_date=start_date, interval=interval
        )



import json
import os
import tempfile
import fnmatch

from typing import Any, Callable, Dict, List, Optional, Union
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from tuf.repository_tool import METADATA_DIRECTORY_NAME
from taf.git import GitRepository
from taf.repository_tool import (
    Repository as TAFRepository,
    get_role_metadata_path,
    get_target_path,
)


class AuthenticationRepository(GitRepository, TAFRepository):

    LAST_VALIDATED_FILENAME = "last_validated_commit"
    TEST_REPO_FLAG_FILE = "test-auth-repo"
    SCRIPTS_PATH = "scripts"

    _conf_dir = None
    _dependencies: Dict = {}

    def __init__(
        self,
        library_dir: Optional[str] = None,
        name: Optional[str] = None,
        urls: Optional[List[str]] = None,
        custom: Optional[Dict] = None,
        default_branch: Optional[str] = None,
        allow_unsafe: Optional[bool] = False,
        conf_directory_root: Optional[str] = None,
        out_of_band_authentication: Optional[str] = None,
        path: Optional[Union[Path, str]] = None,
        alias: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """
        Args:
          library_dir (Path or str): path to the library root. This is a directory which contains other repositories,
          whose full path is determined by appending their names to the library dir. A repository can be
          instantiated by specifying library dir and name, or by the full path.
          name (str): repository's name, which is appended to the library dir to form the full path.
          Must be in the namespace/name format. If library_dir is specified, name must be specified too.
          path (Path): repository's full filesystem path, which can be specified instead of name and library dir
          urls (list): repository's urls
          custom (dict): a dictionary containing other data
          default_branch (str): repository's default branch, automatically determined if not specified
          out_of_band_authentication (str): manually specified initial commit
          alias: Repository's alias, which will be used in logging statements to reference it
        """
        super().__init__(
            library_dir,
            name,
            urls,
            custom,
            default_branch,
            allow_unsafe,
            path,
            alias,
            *args,
            **kwargs,
        )

        if conf_directory_root is None:
            conf_directory_root_path = Path(self.path).parent
        else:
            conf_directory_root_path = Path(conf_directory_root)

        self.conf_directory_root = conf_directory_root_path.resolve()
        self.out_of_band_authentication = out_of_band_authentication

    # TODO rework conf_dir

    def to_json_dict(self) -> Dict:
        """Returns a dictionary mapping all attributes to their values"""
        data = super().to_json_dict()
        data.update(
            {
                "conf_directory_root": str(self.conf_directory_root),
                "out_of_band_authentication": self.out_of_band_authentication,
                "dependencies": self.dependencies,
            }
        )
        return data

    @property
    def conf_dir(self) -> str:
        """
        Returns location of the directory which stores the authentication repository's
        configuration files. That is, the last validated commit.
        Create the directory if it does not exist.
        """
        # the repository's name consists of the namespace and name (namespace/name)
        # the configuration directory should be _name
        if self._conf_dir is None:
            last_dir = os.path.basename(os.path.normpath(self.path))
            conf_path = self.conf_directory_root / f"_{last_dir}"
            conf_path.mkdir(parents=True, exist_ok=True)
            self._conf_dir = str(conf_path)
        return self._conf_dir

    @property
    def certs_dir(self) -> str:
        certs_dir = Path(self.path, "certs")
        return str(certs_dir)

    @property
    def dependencies(self) -> Dict:
        return self._dependencies

    @dependencies.setter
    def dependencies(self, value):
        self._dependencies = value

    @property
    def is_test_repo(self, branch: Optional[str] = None) -> bool:
        try:
            if branch is not None:
                commit = self.top_commit_of_branch(branch)
            else:
                commit = self.head_commit_sha()
            targets_data = self.get_metadata("targets", commit)
            if targets_data is None:
                return False
            return self.TEST_REPO_FLAG_FILE in targets_data["signed"]["targets"]
        except Exception as e:
            self._log_debug(f"Could not get targets.json metadata: {e}")
            return False

    @property
    def last_validated_commit(self) -> Optional[str]:
        """
        Return the last validated commit of the authentication repository
        """
        try:
            return Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).read_text().strip()
        except FileNotFoundError:
            return None

    @property
    def log_prefix(self) -> str:
        if self.alias:
            return f"{self.alias}: "
        return f"Auth repo {self.name}: "

    def get_target(self, target_name, commit=None, safely=True) -> Optional[Dict]:
        if commit is None:
            commit = self.head_commit_sha()
        if commit is None:
            return None
        target_path = get_target_path(target_name)
        if safely:
            return self.safely_get_json(commit, target_path)
        else:
            return self.get_json(commit, target_path)

    def get_metadata(
        self, role: str, commit: Optional[str] = None, safely: bool = True
    ) -> Optional[Dict]:
        if commit is None:
            commit = self.head_commit_sha()
        if commit is None:
            return None
        metadata_path = get_role_metadata_path(role)
        if safely:
            return self.safely_get_json(commit, metadata_path)
        else:
            return self.get_json(commit, metadata_path)

    def is_commit_authenticated(self, target_name: str, commit: str) -> bool:
        """Checks if passed commit is ever authenticated for given target name."""
        for auth_commit in self.all_commits_on_branch(reverse=False):
            target = self.get_target(target_name, auth_commit)
            if target is None:
                continue
            try:
                if target["commit"] == commit:
                    return True
            except TypeError:
                continue
        return False

    @contextmanager
    def repository_at_revision(self, commit: str):
        """
        Context manager which makes sure that TUF repository is instantiated
        using metadata files at the specified revision. Creates a temp directory
        and metadata files inside it. Deleted the temp directory when no longer
        needed.
        """
        tuf_repository = self._tuf_repository
        with tempfile.TemporaryDirectory() as temp_dir:
            metadata_files = self.list_files_at_revision(
                commit, METADATA_DIRECTORY_NAME
            )
            Path(temp_dir, METADATA_DIRECTORY_NAME).mkdir(parents=True)
            for file_name in metadata_files:
                path = Path(temp_dir, METADATA_DIRECTORY_NAME, file_name)
                with open(path, "w") as f:
                    data = self.get_json(
                        commit, f"{METADATA_DIRECTORY_NAME}/{file_name}"
                    )
                    json.dump(data, f)
            self._load_tuf_repository(temp_dir)
            yield
            self._tuf_repository = tuf_repository

    def set_last_validated_commit(self, commit: str):
        """
        Set the last validated commit of the authentication repository
        """
        self._log_debug(f"setting last validated commit to: {commit}")
        Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).write_text(commit)

    def targets_data_by_auth_commits(
        self,
        commits: List[str],
        target_repos: Optional[List[str]] = None,
        custom_fns: Optional[Dict[str, Callable]] = None,
        default_branch: Optional[str] = None,
        excluded_target_globs: Optional[List[str]] = None,
    ) -> Dict[str, Dict[str, Dict[str, Any]]]:
        """
        Return a dictionary where each target repository has associated authentication commits,
        and for each authentication commit, there's a dictionary of the branch, commit and custom data.

        {
            'target_repo1': {
                'auth_commit1': {'branch': 'branch1', 'commit': 'commit1', 'custom': {}},
                'auth_commit2': {'branch': 'branch1', 'commit': 'commit2', 'custom': {}},
                ...
            },
            'target_repo2': {
                ...
            },
            ...
        }

        """
        repositories_commits: Dict[str, Dict[str, Dict[str, Any]]] = {}
        targets = self.targets_at_revisions(
            *commits, target_repos=target_repos, default_branch=default_branch
        )
        excluded_target_globs = excluded_target_globs or []
        for commit in commits:
            for target_path, target_data in targets[commit].items():
                if any(
                    fnmatch.fnmatch(target_path, excluded_target_glob)
                    for excluded_target_glob in excluded_target_globs
                ):
                    continue

                target_branch = target_data.get("branch")
                target_commit = target_data.get("commit")
                target_data.setdefault("custom", {})
                if custom_fns is not None and target_path in custom_fns:
                    target_data["custom"].update(custom_fns[target_path](target_commit))

                repositories_commits.setdefault(target_path, {})[commit] = {
                    "branch": target_branch,
                    "commit": target_commit,
                    "custom": target_data.get("custom"),
                }

        self._log_debug(
            f"new commits per repositories according to target files: {repositories_commits}"
        )
        return repositories_commits

    def sorted_commits_and_branches_per_repositories(
        self,
        commits: List[str],
        target_repos: Optional[List[str]] = None,
        custom_fns: Optional[Dict[str, Callable]] = None,
        default_branch: Optional[str] = None,
        excluded_target_globs: Optional[List[str]] = None,
    ) -> Dict[str, Dict[str, List[Dict[str, Any]]]]:
        """Return a dictionary consisting of branches and commits belonging
        to it for every target repository:
        {
            target_repo1: {
                branch1: [
                    {"commit": commit1, "custom": {...}, "auth_commit": auth_commit1},
                    {"commit": commit2, "custom": {...}, "auth_commit": auth_commit2},
                    {"commit": commit3, "custom": {...}, "auth_commit": auth_commit3}
                ],
                branch2: [
                    {"commit": commit4, "custom": {...}, "auth_commit": auth_commit4},
                    {"commit": commit5, "custom": {...}, "auth_commit": auth_commit5}
                ]
            },
            target_repo2: {
                branch1: [
                    {"commit": commit6, "custom": {...}, "auth_commit": auth_commit6}
                ],
                branch2: [
                    {"commit": commit7, "custom": {...}, "auth_commit": auth_commit7}
                ]
            }
        }
        Keep in mind that targets metadata
        file is not updated everytime something is committed to the authentication repo.
        """
        repositories_commits: Dict = defaultdict(dict)
        targets = self.targets_at_revisions(
            *commits, target_repos=target_repos, default_branch=default_branch
        )
        previous_commits: Dict = {}
        skipped_targets = []
        excluded_target_globs = excluded_target_globs or []
        for commit in commits:
            for target_path, target_data in targets[commit].items():
                if target_path in skipped_targets:
                    continue
                if any(
                    fnmatch.fnmatch(target_path, excluded_target_glob)
                    for excluded_target_glob in excluded_target_globs
                ):
                    skipped_targets.append(target_path)
                    continue
                target_branch = target_data.get("branch")
                target_commit = target_data.get("commit")
                previous_data = previous_commits.get(target_path)
                target_data.setdefault("custom", {})
                if (
                    previous_data is None
                    or (target_commit, target_branch) != previous_data
                ):
                    if custom_fns is not None and target_path in custom_fns:
                        target_data["custom"].update(
                            custom_fns[target_path](target_commit)
                        )

                    repositories_commits[target_path].setdefault(
                        target_branch, []
                    ).append(
                        {
                            "commit": target_commit,
                            "custom": target_data.get("custom"),
                            "auth_commit": commit,
                        }
                    )
                previous_commits[target_path] = (target_commit, target_branch)
        self._log_debug(
            f"new commits per repositories according to target files: {repositories_commits}"
        )
        return repositories_commits

    def targets_at_revisions(self, *commits, target_repos=None, default_branch=None):
        targets = defaultdict(dict)
        if default_branch is None:
            default_branch = self.default_branch
        previous_metadata = []
        new_files = []
        for commit in commits:
            # repositories.json might not exit, if the current commit is
            # the initial commit
            repositories_at_revision = self.safely_get_json(
                commit, get_target_path("repositories.json")
            )
            if repositories_at_revision is None:
                continue
            repositories_at_revision = repositories_at_revision["repositories"]

            current_metadata = self.list_files_at_revision(
                commit, METADATA_DIRECTORY_NAME
            )
            new_files = [
                metadata_path
                for metadata_path in current_metadata
                if metadata_path not in previous_metadata
            ]
            previous_metadata = current_metadata
            if len(new_files):
                with self.repository_at_revision(commit):
                    roles_at_revision = self.get_all_targets_roles()

            for role_name in roles_at_revision:
                # targets metadata files corresponding to the found roles must exist
                targets_at_revision = self.safely_get_json(
                    commit, get_role_metadata_path(role_name)
                )
                if targets_at_revision is None:
                    continue
                targets_at_revision = targets_at_revision["signed"]["targets"]

                for target_path in targets_at_revision:
                    if target_path not in repositories_at_revision:
                        # we only care about repositories
                        continue
                    if target_repos is not None and target_path not in target_repos:
                        # if specific target repositories are specified, skip all other
                        # repositories
                        continue
                    target_content = self.safely_get_json(
                        commit, get_target_path(target_path)
                    )
                    if target_content is not None:
                        target_commit = target_content.pop("commit")
                        target_branch = target_content.pop("branch", default_branch)
                        targets[commit][target_path] = {
                            "branch": target_branch,
                            "commit": target_commit,
                            "custom": target_content,
                        }
        return targets



from logging import DEBUG, INFO
from typing import Dict, List, Optional, Union
from functools import partial
from logdecorator import log_on_end, log_on_start
from tuf.repository_tool import Repository as TUFRepository, Targets
from taf.exceptions import TAFError
from taf.models.types import RolesIterator
from tuf.repository_tool import Metadata
from taf import YubikeyMissingLibrary
from taf.keys import get_key_name
from taf.auth_repo import AuthenticationRepository
from taf.constants import YUBIKEY_EXPIRATION_DATE
from taf.repository_tool import (
    MAIN_ROLES,
    Repository,
    yubikey_signature_provider,
)
from taf.models.types import Role
from taf.log import taf_logger

ykman_installed = True
try:
    import taf.yubikey as yk
except ImportError:
    yk = YubikeyMissingLibrary()  # type: ignore


@log_on_start(INFO, "Creating delegations", logger=taf_logger)
@log_on_end(DEBUG, "Finished creating delegations", logger=taf_logger)
def create_delegations(
    role: Role,
    repository: AuthenticationRepository,
    verification_keys: Dict,
    signing_keys: Dict,
    existing_roles: Optional[List[str]] = None,
) -> None:
    """
    Initialize new delegated target roles, update authentication repository object

    Arguments:
        role: Targets main role or a delegated role
        repository: Authentication repository.
        verification_keys: A dictionary containing mappings of role names to their verification (public) keys.
        signing_keys: A dictionary containing mappings of role names to their signing (private) keys.
        existing_roles: A list of already initialized roles.

    Side Effects:
        Updates authentication repository object

    Returns:
        None
    """
    if existing_roles is None:
        existing_roles = []
    skip_top_role = role.name == "targets"
    for delegated_role in RolesIterator(role, skip_top_role=skip_top_role):
        parent_role_obj = _role_obj(delegated_role.parent.name, repository)
        if not isinstance(parent_role_obj, Targets):
            raise TAFError(
                f"Could not find parent targets role of role {delegated_role}"
            )
        if delegated_role.name in existing_roles:
            taf_logger.info(f"Role {delegated_role.name} already set up.")
            continue
        paths = delegated_role.paths
        roles_verification_keys = verification_keys[delegated_role.name]
        # if yubikeys are used for signing, signing keys are not loaded
        roles_signing_keys = signing_keys.get(delegated_role.name)
        parent_role_obj.delegate(
            delegated_role.name,
            roles_verification_keys,
            paths,
            threshold=delegated_role.threshold,
            terminating=delegated_role.terminating,
        )
        setup_role(
            delegated_role,
            repository,
            roles_verification_keys,
            roles_signing_keys,
            parent=parent_role_obj,
        )


@log_on_start(DEBUG, "Finding roles of key", logger=taf_logger)
def get_roles_and_paths_of_key(
    public_key: Dict,
    repository: AuthenticationRepository,
):
    roles = repository.find_associated_roles_of_key(public_key)
    roles_with_paths: Dict = {role: {} for role in roles}
    for role in roles:
        if role not in MAIN_ROLES:
            roles_with_paths[role] = repository.get_role_paths(role)
    return roles_with_paths


@log_on_start(INFO, "Setting up role {role.name:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished setting up role {role.name:s}", logger=taf_logger)
def setup_role(
    role: Role,
    repository: Repository,
    verification_keys: Dict,
    signing_keys: Optional[Dict] = None,
    parent: Optional[Targets] = None,
) -> None:
    """
    Initialize a new role, add signing and verification keys.
    """
    role_obj = _role_obj(role.name, repository, parent)
    role_obj.threshold = role.threshold
    if not role.is_yubikey:
        if verification_keys is None or signing_keys is None:
            raise TAFError(f"Cannot setup role {role.name}. Keys not specified")
        for public_key, private_key in zip(verification_keys, signing_keys):
            role_obj.add_verification_key(public_key)
            role_obj.load_signing_key(private_key)
    else:
        yubikeys = role.yubikeys
        if yubikeys is None:
            yubikeys = [
                get_key_name(role.name, count, role.number)
                for count in range(role.number)
            ]
        for key_name, key in zip(yubikeys, verification_keys):
            role_obj.add_verification_key(key, expires=YUBIKEY_EXPIRATION_DATE)
            # check if yubikey loaded
            if yk.get_key_serial_by_id(key_name):
                role_obj.add_external_signature_provider(
                    key, partial(yubikey_signature_provider, key_name, key["keyid"])
                )


def _role_obj(
    role: str,
    repository: Union[Repository, TUFRepository],
    parent: Optional[Targets] = None,
) -> Metadata:
    """
    Return role TUF object based on its name
    """
    if isinstance(repository, Repository):
        tuf_repository = repository._repository
    else:
        tuf_repository = repository
    if role == "targets":
        return tuf_repository.targets
    elif role == "snapshot":
        return tuf_repository.snapshot
    elif role == "timestamp":
        return tuf_repository.timestamp
    elif role == "root":
        return tuf_repository.root
    else:
        # return delegated role
        if parent is None:
            return tuf_repository.targets(role)
        return parent(role)



from collections import defaultdict
from enum import Enum
import functools
from logging import DEBUG, INFO
from pathlib import Path
import shutil
import tempfile
from typing import Any, Dict, List, Optional
from attr import attrs, define, field
from taf.git import GitError
from logdecorator import log_on_end, log_on_start
from taf.git import GitRepository

import taf.settings as settings
import taf.repositoriesdb as repositoriesdb
from taf.auth_repo import AuthenticationRepository
from taf.exceptions import (
    MissingInfoJsonError,
    RepositoryNotCleanError,
    UpdateFailedError,
)
from taf.updater.handlers import GitUpdater
from taf.updater.lifecycle_handlers import Event
from taf.updater.types.update import OperationType, UpdateType
from taf.utils import on_rm_error
from taf.log import taf_logger
from tuf.ngclient.updater import Updater
from tuf.repository_tool import TARGETS_DIRECTORY_NAME


EXPIRED_METADATA_ERROR = "ExpiredMetadataError"
PROTECTED_DIRECTORY_NAME = "protected"
INFO_JSON_PATH = f"{TARGETS_DIRECTORY_NAME}/{PROTECTED_DIRECTORY_NAME}/info.json"


class UpdateStatus(Enum):
    SUCCESS = 1
    PARTIAL = 2
    FAILED = 3


class RunMode(Enum):
    UPDATE = 1
    LOCAL_VALIDATION = 2
    ALL = 3


@define
class UpdateState:
    auth_commits_since_last_validated: List[Any] = field(factory=list)
    existing_repo: bool = field(default=False)
    is_test_repo: bool = field(default=False)
    update_status: UpdateStatus = field(default=None)
    update_successful: bool = field(default=False)
    event: Optional[str] = field(default=None)
    users_auth_repo: Optional["AuthenticationRepository"] = field(default=None)
    validation_auth_repo: Optional["AuthenticationRepository"] = field(default=None)
    auth_repo_name: Optional[str] = field(default=None)
    errors: Optional[List[Exception]] = field(default=None)
    targets_data: Dict[str, Any] = field(factory=dict)
    last_validated_commit: str = field(factory=str)
    target_repositories: Dict[str, "GitRepository"] = field(factory=dict)
    cloned_target_repositories: List["GitRepository"] = field(factory=list)
    target_branches_data_from_auth_repo: Dict = field(factory=dict)
    targets_data_by_auth_commits: Dict = field(factory=dict)
    old_heads_per_target_repos_branches: Dict[str, Dict[str, str]] = field(factory=dict)
    fetched_commits_per_target_repos_branches: Dict[str, Dict[str, List[str]]] = field(
        factory=dict
    )
    validated_commits_per_target_repos_branches: Dict[str, Dict[str, str]] = field(
        factory=dict
    )
    additional_commits_per_target_repos_branches: Dict[
        str, Dict[str, List[str]]
    ] = field(factory=dict)
    validated_auth_commits: List[str] = field(factory=list)


@attrs
class UpdateOutput:
    event: str = field()
    users_auth_repo: Any = field()
    auth_repo_name: str = field()
    commits_data: Dict[str, Any] = field()
    error: Optional[Exception] = field(default=None)
    targets_data: Dict[str, Any] = field(factory=dict)


def cleanup_decorator(pipeline_function):
    @functools.wraps(pipeline_function)
    def wrapper(self, *args, **kwargs):
        try:
            result = pipeline_function(self, *args, **kwargs)
            return result
        finally:
            if (
                self.state.event == Event.FAILED
                and not self.state.existing_repo
                and self.state.users_auth_repo is not None
            ):
                shutil.rmtree(self.state.users_auth_repo.path, onerror=on_rm_error)
                shutil.rmtree(self.state.users_auth_repo.conf_dir)

    return wrapper


def combine_statuses(current_status, new_status):
    """
    Combine the current overall status with the new step status.
    """
    if current_status is None:
        return new_status
    if new_status == UpdateStatus.FAILED:
        return UpdateStatus.FAILED
    elif new_status == UpdateStatus.PARTIAL:
        if current_status != UpdateStatus.FAILED:
            return UpdateStatus.PARTIAL
    return current_status


def format_commit(commit):
    return commit[:10]


class Pipeline:
    def __init__(self, steps, run_mode):
        self.steps = steps
        self.current_step = None
        self.run_mode = run_mode

    def run(self):
        self.state.errors = []
        for step, step_run_mode in self.steps:
            try:
                if step_run_mode == RunMode.ALL or step_run_mode == self.run_mode:
                    self.current_step = step
                    update_status = step()
                    combined_status = combine_statuses(
                        self.state.update_status, update_status
                    )
                    self.state.update_status = combined_status
                    if combined_status == UpdateStatus.FAILED:
                        message = "\n".join(str(error) for error in self.state.errors)
                        raise UpdateFailedError(message)

            except Exception as e:
                self.handle_error(e)
                break
        self.set_output()

    def handle_error(self, e):
        if self.state.auth_repo_name is not None:
            taf_logger.error(
                "An error occurred while updating repository {} while running step {}: {}",
                self.state.auth_repo_name,
                self.current_step.__name__,
                str(e),
            )
        else:
            taf_logger.error(
                "An error occurred while updating authentication repository while running step {}: {}",
                self.current_step.__name__,
                str(e),
            )

    def set_output(self):
        pass


class AuthenticationRepositoryUpdatePipeline(Pipeline):
    def __init__(
        self,
        operation,
        url,
        auth_path,
        library_dir,
        update_from_filesystem,
        expected_repo_type,
        target_repo_classes,
        target_factory,
        only_validate,
        validate_from_commit,
        conf_directory_root,
        out_of_band_authentication,
        checkout,
        excluded_target_globs,
    ):

        super().__init__(
            steps=[
                (self.clone_remote_and_run_tuf_updater, RunMode.ALL),
                (self.validate_out_of_band_and_update_type, RunMode.ALL),
                (self.check_if_local_auth_repo_clean, RunMode.UPDATE),
                (self.clone_or_fetch_users_auth_repo, RunMode.UPDATE),
                (self.load_target_repositories, RunMode.ALL),
                (self.check_if_repositories_on_disk, RunMode.LOCAL_VALIDATION),
                (self.check_if_local_target_repositories_clean, RunMode.UPDATE),
                (self.clone_target_repositories_if_not_on_disk, RunMode.UPDATE),
                (self.determine_start_commits, RunMode.ALL),
                (self.get_targets_data_from_auth_repo, RunMode.ALL),
                (self.get_target_repositories_commits, RunMode.ALL),
                (self.validate_target_repositories, RunMode.ALL),
                (
                    self.validate_and_set_additional_commits_of_target_repositories,
                    RunMode.ALL,
                ),
                (self.merge_commits, RunMode.UPDATE),
                (self.set_target_repositories_data, RunMode.UPDATE),
                (self.print_additional_commits, RunMode.ALL),
            ],
            run_mode=RunMode.LOCAL_VALIDATION if only_validate else RunMode.UPDATE,
        )

        self.operation = operation
        self.url = url
        self.library_dir = library_dir
        self.auth_path = auth_path
        self.update_from_filesystem = update_from_filesystem
        self.expected_repo_type = expected_repo_type
        self.target_repo_classes = target_repo_classes
        self.target_factory = target_factory
        self.only_validate = only_validate
        self.validate_from_commit = validate_from_commit
        self.conf_directory_root = conf_directory_root
        self.out_of_band_authentication = out_of_band_authentication
        self.checkout = checkout
        self.excluded_target_globs = excluded_target_globs
        self.state = UpdateState()
        self.state.targets_data = {}
        self._output = None

    @property
    def output(self):
        if not self._output:
            raise ValueError(
                "Pipeline has not been run yet. Please run the pipeline first."
            )
        return self._output

    @log_on_start(
        INFO, "Cloning repository and running TUF updater...", logger=taf_logger
    )
    @cleanup_decorator
    def clone_remote_and_run_tuf_updater(self):
        settings.update_from_filesystem = self.update_from_filesystem
        settings.conf_directory_root = self.conf_directory_root

        if self.operation == OperationType.CLONE_OR_UPDATE:
            if (
                self.auth_path is not None
                and AuthenticationRepository(path=self.auth_path).is_git_repository
            ):
                self.operation = OperationType.UPDATE
            else:
                self.operation = OperationType.CLONE

        # set last validated commit before running the updater
        # this last validated commit is read from the settings
        if self.operation == OperationType.CLONE:
            settings.last_validated_commit = None
        elif not settings.overwrite_last_validated_commit:
            users_auth_repo = AuthenticationRepository(path=self.auth_path)
            last_validated_commit = users_auth_repo.last_validated_commit
            settings.last_validated_commit = last_validated_commit

        try:
            self.state.auth_commits_since_last_validated = None

            validation_repo = _clone_validation_repo(self.url)

            # check if auth path is provided and if that is not the case
            # check if info.json exists. info.json will be read after validation
            # at revision determined by the last validated commit
            # but we do not want the user to have to wait for the validation to be over
            # before raising an error because info.json is missing
            top_commit_of_validation_repo = validation_repo.top_commit_of_branch(
                validation_repo.default_branch
            )
            auth_repo_name = None
            git_updater = None

            if self.auth_path:
                auth_repo_name = GitRepository(path=self.auth_path).name
                self.state.auth_repo_name = auth_repo_name
            else:
                auth_repo_name = _get_repository_name_raise_error_if_not_defined(
                    validation_repo, top_commit_of_validation_repo
                )

            git_updater = GitUpdater(self.url, self.library_dir, validation_repo.name)
            last_validated_remote_commit, error = _run_tuf_updater(
                git_updater, auth_repo_name
            )
            if last_validated_remote_commit is None and error is not None:
                raise error

            # having validated the repository, read info.json from the last
            # valid commit if it is not the same as the most recent commit
            if self.state.auth_repo_name is None:
                if top_commit_of_validation_repo != last_validated_remote_commit:
                    self.state.auth_repo_name = (
                        _get_repository_name_raise_error_if_not_defined(
                            validation_repo, last_validated_remote_commit
                        )
                    )
                else:
                    self.state.auth_repo_name = auth_repo_name

            self.state.users_auth_repo = AuthenticationRepository(
                self.library_dir,
                self.state.auth_repo_name,
                urls=[self.url],
            )

            self.state.existing_repo = self.state.users_auth_repo.is_git_repository_root
            self._validate_operation_type()
            self.state.validation_auth_repo = git_updater.validation_auth_repo
            self.state.is_test_repo = self.state.validation_auth_repo.is_test_repo

            if self.operation == OperationType.UPDATE:
                self._validate_last_validated_commit(settings.last_validated_commit)

            # used for testing purposes
            if settings.overwrite_last_validated_commit:
                self.state.last_validated_commit = settings.last_validated_commit
            else:
                self.state.last_validated_commit = (
                    self.state.users_auth_repo.last_validated_commit
                )

            if error is None:
                self.state.auth_commits_since_last_validated = list(git_updater.commits)
                taf_logger.info(
                    "Successfully validated authentication repository {}",
                    self.state.auth_repo_name,
                )
                self.state.event = (
                    Event.CHANGED
                    if len(self.state.auth_commits_since_last_validated) > 1
                    else Event.UNCHANGED
                )
                return UpdateStatus.SUCCESS
            else:
                validated_commits_since_last_validated = []
                for commit in git_updater.commits:
                    validated_commits_since_last_validated.append(commit)
                    if commit == last_validated_remote_commit:
                        break

                self.state.auth_commits_since_last_validated = (
                    validated_commits_since_last_validated
                )
                self.state.errors.append(error)
                self.state.event = Event.PARTIAL
                return UpdateStatus.PARTIAL

        except Exception as e:
            self.state.errors.append(e)
            self.state.users_auth_repo = None

            if self.state.auth_repo_name is not None:
                self.state.users_auth_repo = AuthenticationRepository(
                    self.library_dir,
                    self.state.auth_repo_name,
                    urls=[self.url],
                    conf_directory_root=self.conf_directory_root,
                )
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED
        finally:
            # always clean up repository updater
            if git_updater is not None:
                git_updater.cleanup()

    def _validate_operation_type(self):
        if self.operation == OperationType.CLONE and self.state.existing_repo:
            raise UpdateFailedError(
                f"Destination path {self.state.users_auth_repo.path} already exists and is not an empty directory. Run `taf repo update` to update it."
            )
        elif self.operation == OperationType.UPDATE and not self.state.existing_repo:
            raise UpdateFailedError(
                f"{self.state.users_auth_repo.path} is not a Git repository. Run `taf repo clone` instead"
            )

    @log_on_start(
        INFO, "Validating out of band commit and update type", logger=taf_logger
    )
    def validate_out_of_band_and_update_type(self):
        # this is the repository cloned inside the temp directory
        # we validate it before updating the actual authentication repository
        try:
            if (
                self.out_of_band_authentication is not None
                and self.state.users_auth_repo.last_validated_commit is None
                and self.state.auth_commits_since_last_validated[0]
                != self.out_of_band_authentication
            ):
                raise UpdateFailedError(
                    f"First commit of repository {self.state.auth_repo_name} does not match "
                    "out of band authentication commit"
                )

            if self.expected_repo_type != UpdateType.EITHER:
                # check if the repository being updated is a test repository
                if (
                    self.state.is_test_repo
                    and self.expected_repo_type != UpdateType.TEST
                ):
                    raise UpdateFailedError(
                        f"Repository {self.state.users_auth_repo.name} is a test repository. "
                        'Call update with "--expected-repo-type" test to update a test '
                        "repository"
                    )
                elif (
                    not self.state.is_test_repo
                    and self.expected_repo_type == UpdateType.TEST
                ):
                    raise UpdateFailedError(
                        f"Repository {self.state.users_auth_repo.name} is not a test repository,"
                        ' but update was called with the "--expected-repo-type" test'
                    )
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    def _validate_last_validated_commit(self, last_validated_commit):
        branch = self.state.users_auth_repo.default_branch
        users_head_sha = self.state.users_auth_repo.top_commit_of_branch(branch)
        if last_validated_commit != users_head_sha:
            # if a user committed something to the repo or manually pulled the changes
            # last_validated_commit will no longer match the top commit, but the repository
            # might still be completely valid
            # committing without pushing is not valid
            # user_head_sha should be newer than last validated commit
            commits_since = self.state.users_auth_repo.all_commits_since_commit(
                since_commit=last_validated_commit, branch=branch
            )
            if users_head_sha not in commits_since:
                msg = f"Top commit of repository {self.state.users_auth_repo.name} {users_head_sha} and is not equal to or newer than last successful commit"
                taf_logger.error(msg)
                raise UpdateFailedError(msg)

    @log_on_start(
        INFO,
        "Checking if local auth repo is clean...",
        logger=taf_logger,
    )
    def check_if_local_auth_repo_clean(self):
        try:
            if self.state.existing_repo:
                if self.state.users_auth_repo.something_to_commit():
                    raise RepositoryNotCleanError(self.state.users_auth_repo.name)
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED
        return UpdateStatus.SUCCESS

    @log_on_start(
        INFO,
        "Cloning or updating user's authentication repository...",
        logger=taf_logger,
    )
    def clone_or_fetch_users_auth_repo(self):
        # fetch the latest commit or clone the repository without checkout
        # do not merge before targets are validated as well
        try:
            if self.state.existing_repo:
                self.state.users_auth_repo.fetch(fetch_all=True)
            else:
                self.state.users_auth_repo.clone()
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED
        return UpdateStatus.SUCCESS

    @log_on_start(DEBUG, "Loading target repositories", logger=taf_logger)
    def load_target_repositories(self):
        try:
            repositoriesdb.load_repositories(
                self.state.users_auth_repo,
                repo_classes=self.target_repo_classes,
                factory=self.target_factory,
                library_dir=self.library_dir,
                commits=self.state.auth_commits_since_last_validated,
                only_load_targets=True,
                excluded_target_globs=self.excluded_target_globs,
            )
            self.state.target_repositories = (
                repositoriesdb.get_deduplicated_repositories(
                    self.state.users_auth_repo,
                    self.state.auth_commits_since_last_validated[-1::],
                )
            )
            return UpdateStatus.SUCCESS
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    @log_on_start(
        INFO,
        "Checking if all target repositories are already on disk...",
        logger=taf_logger,
    )
    def check_if_repositories_on_disk(self):
        try:
            for repository in self.state.target_repositories.values():
                if not repository.is_git_repository_root:
                    is_git_repository = repository.is_git_repository_root
                    if not is_git_repository:
                        if self.only_validate:
                            self.state.targets_data = {}
                            msg = f"{repository.name} not on disk. Please run update to clone the repositories."
                            taf_logger.error(msg)
                            raise UpdateFailedError(msg)
            return UpdateStatus.SUCCESS
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    @log_on_start(
        INFO,
        "Checking if all existing target repositories are clean...",
        logger=taf_logger,
    )
    def check_if_local_target_repositories_clean(self):
        try:
            for repository in self.state.target_repositories.values():
                if repository.is_git_repository_root:
                    if repository.something_to_commit():
                        raise RepositoryNotCleanError(repository.name)
            return UpdateStatus.SUCCESS
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    @log_on_start(
        INFO, "Cloning target repositories which are not on disk...", logger=taf_logger
    )
    @log_on_end(INFO, "Finished cloning target repositories", logger=taf_logger)
    def clone_target_repositories_if_not_on_disk(self):
        try:
            self.state.cloned_target_repositories = []
            for repository in self.state.target_repositories.values():
                is_git_repository = repository.is_git_repository_root
                if not is_git_repository:
                    repository.clone(no_checkout=True)
                    self.state.cloned_target_repositories.append(repository)
            return UpdateStatus.SUCCESS
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    @log_on_start(
        INFO, "Validating initial state of target repositories...", logger=taf_logger
    )
    @log_on_end(
        INFO,
        "Checking initial state of repositories",
        logger=taf_logger,
    )
    def determine_start_commits(self):
        try:
            self.state.targets_data_by_auth_commits = (
                self.state.users_auth_repo.targets_data_by_auth_commits(
                    self.state.auth_commits_since_last_validated
                )
            )
            self.state.old_heads_per_target_repos_branches = defaultdict(dict)
            is_initial_state_in_sync = True
            # if last validated commit was not manually modified (set to a newer commit)
            # target repositories data that is extracted to them (commit and branch)
            # should be present in the local repository
            # if the local repository was manually modified (say, something was committed)
            # we still expect the last validated target commit to exist
            # and the remaining commits will be validated afterwards
            # if the last validated target commit does not exist, start the validation from scratch
            if self.state.last_validated_commit is not None:
                for repository in self.state.target_repositories.values():
                    if repository.name not in self.state.targets_data_by_auth_commits:
                        continue
                    self.state.old_heads_per_target_repos_branches[repository.name] = {}
                    last_validated_repository_commits_data = (
                        self.state.targets_data_by_auth_commits[repository.name].get(
                            self.state.last_validated_commit, {}
                        )
                    )

                    if last_validated_repository_commits_data:
                        # if this is not set, it means that the repository did not exist in this revision
                        if repository in self.state.cloned_target_repositories:
                            is_initial_state_in_sync = False
                            break
                        current_branch = last_validated_repository_commits_data.get(
                            "branch", repository.default_branch
                        )
                        last_validated_commit = last_validated_repository_commits_data[
                            "commit"
                        ]

                        branch_exists = repository.branch_exists(
                            current_branch, include_remotes=False
                        )
                        if not branch_exists:
                            is_initial_state_in_sync = False
                            break
                        top_commit_of_branch = repository.top_commit_of_branch(
                            current_branch
                        )
                        if top_commit_of_branch != last_validated_commit:
                            # check if top commit is newer (which is fine, it will be validated)
                            # or older, meaning that the authentication repository contains
                            # additional commits, so it would be necessary to find older auth repo
                            # commit and start the validation from there
                            if (
                                current_branch
                                not in repository.branches_containing_commit(
                                    last_validated_commit
                                )
                            ):
                                is_initial_state_in_sync = False
                                break

                        self.state.old_heads_per_target_repos_branches[repository.name][
                            current_branch
                        ] = last_validated_commit

            if not is_initial_state_in_sync:
                taf_logger.info(
                    f"Repository {self.state.users_auth_repo.name}: states of target repositories are not in sync with last validated commit. Starting the update from the beginning"
                )
                self.state.last_validated_commit = None
                self.state.auth_commits_since_last_validated = (
                    self.state.users_auth_repo.all_commits_on_branch(
                        self.state.users_auth_repo.default_branch
                    )
                )
                self.state.targets_data_by_auth_commits = (
                    self.state.users_auth_repo.targets_data_by_auth_commits(
                        self.state.auth_commits_since_last_validated
                    )
                )

            return UpdateStatus.SUCCESS
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    def get_targets_data_from_auth_repo(self):
        repo_branches = {}
        for repo_name, commits_data in self.state.targets_data_by_auth_commits.items():
            branches = set()  # using a set to avoid duplicate branches
            for commit_data in commits_data.values():
                branches.add(commit_data["branch"])
            repo_branches[repo_name] = sorted(list(branches))
        self.state.target_branches_data_from_auth_repo = repo_branches
        return UpdateStatus.SUCCESS

    @log_on_start(DEBUG, "Fetching commits of target repositories", logger=taf_logger)
    def get_target_repositories_commits(self):
        """Returns a list of newly fetched commits belonging to the specified branch."""
        self.state.fetched_commits_per_target_repos_branches = defaultdict(dict)
        for repository in self.state.target_repositories.values():
            if repository.name not in self.state.target_branches_data_from_auth_repo:
                # exists in repositories.json, not target files
                continue
            for branch in self.state.target_branches_data_from_auth_repo[
                repository.name
            ]:
                if repository not in self.state.cloned_target_repositories:
                    if self.only_validate:
                        branch_exists = repository.branch_exists(
                            branch, include_remotes=False
                        )
                        if not branch_exists:
                            self.state.targets_data = {}
                            msg = f"{repository.name} does not contain a local branch named {branch} and cannot be validated. Please update the repositories."
                            taf_logger.error(msg)
                            raise UpdateFailedError(msg)
                    else:
                        repository.fetch(branch=branch)

                old_head = self.state.old_heads_per_target_repos_branches[
                    repository.name
                ].get(branch)
                if old_head is not None:
                    if not self.only_validate:
                        fetched_commits = repository.all_commits_on_branch(
                            branch=f"origin/{branch}"
                        )

                        # if the local branch does not exist (the branch was not checked out locally)
                        # fetched commits will include already validated commits
                        # check which commits are newer that the previous head commit
                        if old_head in fetched_commits:
                            fetched_commits_on_target_repo_branch = fetched_commits[
                                fetched_commits.index(old_head) + 1 : :
                            ]
                        else:
                            fetched_commits_on_target_repo_branch = (
                                repository.all_commits_since_commit(old_head, branch)
                            )
                            for commit in fetched_commits:
                                if commit not in fetched_commits_on_target_repo_branch:
                                    fetched_commits_on_target_repo_branch.append(commit)
                    else:
                        fetched_commits_on_target_repo_branch = (
                            repository.all_commits_since_commit(old_head, branch)
                        )
                    fetched_commits_on_target_repo_branch.insert(0, old_head)
                else:
                    branch_exists = repository.branch_exists(
                        branch, include_remotes=False
                    )
                    if branch_exists:
                        # this happens in the case when last_validated_commit does not exist
                        # we want to validate all commits, so combine existing commits and
                        # fetched commits
                        fetched_commits_on_target_repo_branch = (
                            repository.all_commits_on_branch(
                                branch=branch, reverse=True
                            )
                        )
                    else:
                        fetched_commits_on_target_repo_branch = []
                    if not self.only_validate:
                        try:
                            fetched_commits = repository.all_commits_on_branch(
                                branch=f"origin/{branch}"
                            )

                            # if the local branch does not exist (the branch was not checked out locally)
                            # fetched commits will include already validated commits
                            # check which commits are newer that the previous head commit
                            for commit in fetched_commits:
                                if commit not in fetched_commits_on_target_repo_branch:
                                    fetched_commits_on_target_repo_branch.append(commit)
                        except GitError:
                            pass
                self.state.fetched_commits_per_target_repos_branches[repository.name][
                    branch
                ] = fetched_commits_on_target_repo_branch
        return UpdateStatus.SUCCESS

    @log_on_start(INFO, "Validating target repositories...", logger=taf_logger)
    @log_on_end(INFO, "Validation of target repositories finished", logger=taf_logger)
    def validate_target_repositories(self):
        """
        Breadth-first update of target repositories
        Merge last valid commits at the end of the update
        """
        try:
            # need to be set to old head since that is the last validated target
            self.state.validated_commits_per_target_repos_branches = defaultdict(dict)

            last_validated_data_per_repositories = defaultdict(dict)
            self.state.validated_auth_commits = []
            for auth_commit in self.state.auth_commits_since_last_validated:
                for repository in self.state.target_repositories.values():
                    if repository.name not in self.state.targets_data_by_auth_commits:
                        continue
                    if (
                        auth_commit
                        not in self.state.targets_data_by_auth_commits[repository.name]
                    ):
                        continue
                    current_targets_data = self.state.targets_data_by_auth_commits[
                        repository.name
                    ][auth_commit]

                    current_branch = current_targets_data.get(
                        "branch", repository.default_branch
                    )
                    current_commit = current_targets_data["commit"]

                    if not len(last_validated_data_per_repositories[repository.name]):
                        current_head_commit_and_branch = (
                            self.state.targets_data_by_auth_commits[
                                repository.name
                            ].get(self.state.last_validated_commit, {})
                        )
                        previous_branch = current_head_commit_and_branch.get("branch")
                        previous_commit = current_head_commit_and_branch.get("commit")
                        if previous_commit is not None and previous_branch is None:
                            previous_branch = repository.default_branch
                    else:
                        previous_branch = last_validated_data_per_repositories[
                            repository.name
                        ].get("branch")
                        previous_commit = last_validated_data_per_repositories[
                            repository.name
                        ]["commit"]

                    target_commits_from_target_repo = (
                        self.state.fetched_commits_per_target_repos_branches[
                            repository.name
                        ]
                    )
                    validated_commit = self._validate_current_repo_commit(
                        repository,
                        self.state.users_auth_repo,
                        previous_branch,
                        previous_commit,
                        current_branch,
                        current_commit,
                        target_commits_from_target_repo,
                        auth_commit,
                    )

                    last_validated_data_per_repositories[repository.name] = {
                        "commit": validated_commit,
                        "branch": current_branch,
                    }

                    self.state.validated_commits_per_target_repos_branches[
                        repository.name
                    ].setdefault(current_branch, []).append(validated_commit)

                # commit processed without an error
                self.state.validated_auth_commits.append(auth_commit)
            return UpdateStatus.SUCCESS
        except Exception as e:
            self.state.errors.append(e)
            taf_logger.error(e)
            if len(self.state.validated_auth_commits):
                self.state.event = Event.PARTIAL
                return UpdateStatus.PARTIAL
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    def _validate_current_repo_commit(
        self,
        repository,
        users_auth_repo,
        previous_branch,
        previous_commit,
        current_branch,
        current_commit,
        target_commits_from_target_repo,
        current_auth_commit,
    ):
        target_commits_from_target_repos_on_branch = target_commits_from_target_repo[
            current_branch
        ]
        if previous_commit == current_commit:
            # target not updated in this revision
            return current_commit
        if previous_branch == current_branch:
            # same branch
            current_target_commit = _find_next_value(
                previous_commit, target_commits_from_target_repos_on_branch
            )
        else:
            # next branch
            current_target_commit = target_commits_from_target_repos_on_branch[0]

        if current_target_commit is None:
            # there are commits missing from the target repository
            commit_date = users_auth_repo.get_commit_date(current_auth_commit)
            raise UpdateFailedError(
                f"Failure to validate {users_auth_repo.name} commit {current_auth_commit} committed on {commit_date}: \
data repository {repository.name} was supposed to be at commit {current_commit} \
but commit not on branch {current_branch}"
            )

        if current_commit == current_target_commit:
            return current_target_commit
        if not _is_unauthenticated_allowed(repository):
            commit_date = users_auth_repo.get_commit_date(current_auth_commit)
            raise UpdateFailedError(
                f"Failure to validate {users_auth_repo.name} commit {current_auth_commit} committed on {commit_date}: \
data repository {repository.name} was supposed to be at commit {current_commit} \
but repo was at {current_target_commit}"
            )
        # unauthenticated commits are allowed, try to skip them
        # if commits of the target repositories were swapped, commit which is expected to be found
        # after the current one will be skipped and it won't be found later, so validation will fail
        remaining_commits = target_commits_from_target_repos_on_branch[
            target_commits_from_target_repos_on_branch.index(current_target_commit) :
        ]
        for target_commit in remaining_commits:
            if current_commit == target_commit:
                return target_commit
            taf_logger.debug(
                f"{repository.name}: skipping target commit {target_commit}. Looking for commit {current_commit}"
            )
        commit_date = users_auth_repo.get_commit_date(current_auth_commit)
        raise UpdateFailedError(
            f"Failure to validate {users_auth_repo.name} commit {current_auth_commit} committed on {commit_date}: \
data repository {repository.name} was supposed to be at commit {current_commit} \
but commit not on branch {current_branch}"
        )

    @log_on_start(
        DEBUG,
        "Validating and setting additional commits of target repositories",
        logger=taf_logger,
    )
    def validate_and_set_additional_commits_of_target_repositories(self):
        """
        For target repository and for each branch, extract commits following the last validated commit
        These commits are not invalid. In case of repositories which cannot contain unauthenticated commits
        all of these commits will have to get signed if at least one commits on that branch needs to get signed
        However, no error will get reported if there are commits which have not yet been signed
        In case of repositories which can contain unauthenticated commits, they do not even need to get signed
        """
        # only get additional commits if the validation was complete (not partial, up to a commit)
        self.state.additional_commits_per_target_repos_branches = defaultdict(dict)
        if self.state.update_status != UpdateStatus.SUCCESS:
            return self.state.update_status
        try:
            for repository in self.state.target_repositories.values():
                # this will only include branches that were, at least partially, validated (up until a certain point)
                for (
                    branch,
                    validated_commits,
                ) in self.state.validated_commits_per_target_repos_branches[
                    repository.name
                ].items():
                    last_validated_commit = validated_commits[-1]
                    branch_commits = (
                        self.state.fetched_commits_per_target_repos_branches[
                            repository.name
                        ][branch]
                    )
                    additional_commits = branch_commits[
                        branch_commits.index(last_validated_commit) + 1 :
                    ]
                    if len(additional_commits):
                        if not _is_unauthenticated_allowed(repository):
                            raise UpdateFailedError(
                                f"Target repository {repository.name} does not allow unauthenticated commits, but contains commit(s) {', '.join(additional_commits)} on branch {branch}"
                            )

                        # these commits include all commits newer than last authenticated commit (if unauthenticated commits are allowed)
                        # that does not necessarily mean that the local repository is not up to date with the remote
                        # pull could've been run manually
                        # check where the current local head is
                        branch_current_head = repository.top_commit_of_branch(branch)
                        if branch_current_head in additional_commits:
                            additional_commits = additional_commits[
                                additional_commits.index(branch_current_head) + 1 :
                            ]
                    self.state.additional_commits_per_target_repos_branches[
                        repository.name
                    ][branch] = additional_commits
            return self.state.update_status
        except UpdateFailedError as e:
            self.state.errors.append(e)
            taf_logger.error(e)
            self.state.event = Event.PARTIAL
            return UpdateStatus.PARTIAL
        except Exception as e:
            self.state.errors.append(e)
            taf_logger.error(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    @log_on_start(
        INFO, "Merging commits into target repositories...", logger=taf_logger
    )
    def merge_commits(self):
        """Determines which commits needs to be merged into the specified branch and
        merge it.
        """
        try:
            if self.only_validate:
                return self.state.update_status
            if (
                self.state.update_status == UpdateStatus.FAILED
                and self.operation == OperationType.CLONE
            ):
                # couldn't validate the first new commit
                # there is nothing to merge
                # remove cloned repositories if the initial commit was incorrect
                for repository in self.state.cloned_target_repositories:
                    taf_logger.debug("Removing cloned repository {}", repository.path)
                    shutil.rmtree(repository.path, onerror=on_rm_error)
            else:
                for repository in self.state.target_repositories.values():
                    # this will only include branches that were, at least partially, validated (up until a certain point)
                    for (
                        branch,
                        validated_commits,
                    ) in self.state.validated_commits_per_target_repos_branches[
                        repository.name
                    ].items():
                        last_validated_commit = validated_commits[-1]
                        commit_to_merge = last_validated_commit

                        taf_logger.info(
                            "Repository {}: merging {} into branch {}",
                            repository.name,
                            format_commit(commit_to_merge),
                            branch,
                        )
                        _merge_commit(
                            repository, branch, commit_to_merge, force_revert=True
                        )
                return self.state.update_status
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    def set_target_repositories_data(self):
        try:
            targets_data = {}
            for repo_name, repo in self.state.target_repositories.items():
                targets_data[repo_name] = {"repo_data": repo.to_json_dict()}
                if repo_name not in self.state.targets_data_by_auth_commits:
                    continue
                commits_data = self.state.targets_data_by_auth_commits[repo_name]

                branch_data = defaultdict(dict)

                # Iterate through auth_commits in the specified order
                for auth_commit in self.state.validated_auth_commits:
                    commit_info = commits_data.get(auth_commit)
                    if not commit_info or "branch" not in commit_info:
                        continue

                    branch = commit_info.pop("branch")

                    # Update the before_pull, after_pull, and new values
                    if branch not in branch_data:
                        old_head = self.state.old_heads_per_target_repos_branches.get(
                            repo_name, {}
                        ).get(branch)
                        branch_data[branch]["new"] = [commit_info]
                        branch_data[branch]["after_pull"] = [commit_info]
                        branch_data[branch][
                            "unauthenticated"
                        ] = self.state.additional_commits_per_target_repos_branches.get(
                            repo_name, {}
                        ).get(
                            branch, []
                        )
                        if old_head is not None:
                            branch_data[branch]["before_pull"] = old_head
                    else:
                        branch_data[branch]["new"].append(commit_info)
                        branch_data[branch]["after_pull"] = commit_info

                targets_data[repo_name]["commits"] = branch_data

            self.state.targets_data = targets_data
            return self.state.update_status
        except Exception as e:
            self.state.errors.append(e)
            self.state.event = Event.FAILED
            return UpdateStatus.FAILED

    def set_output(self):
        if self.state.auth_commits_since_last_validated is None:
            commit_before_pull = None
            new_commits = []
            commit_after_pull = None
        else:
            commit_before_pull = (
                self.state.validated_auth_commits[0]
                if self.state.existing_repo and len(self.state.validated_auth_commits)
                else None
            )

            if len(self.state.validated_auth_commits):
                commit_after_pull = self.state.validated_auth_commits[-1]
            else:
                commit_after_pull = None

            if not self.state.existing_repo:
                new_commits = self.state.validated_auth_commits
            else:
                new_commits = (
                    self.state.validated_auth_commits[1:]
                    if len(self.state.validated_auth_commits)
                    else []
                )
        commits_data = {
            "before_pull": commit_before_pull,
            "new": new_commits,
            "after_pull": commit_after_pull,
        }

        if len(self.state.errors):
            message = "\n".join(str(error) for error in self.state.errors)
            error = UpdateFailedError(message)
        else:
            error = None

        self._output = UpdateOutput(
            event=self.state.event,
            users_auth_repo=self.state.users_auth_repo,
            auth_repo_name=self.state.auth_repo_name,
            commits_data=commits_data,
            error=error,
            targets_data=self.state.targets_data,
        )

    def print_additional_commits(self):
        for (
            repo_name,
            branches,
        ) in self.state.additional_commits_per_target_repos_branches.items():
            for branch_name, additional_commits in branches.items():
                if len(additional_commits):
                    formatted_commits = [
                        format_commit(commit) for commit in additional_commits
                    ]
                    taf_logger.info(
                        f"Repository {repo_name}: found commits succeeding the last authenticated commit on branch {branch_name}: {', '.join(formatted_commits)}.\nThese commits were not merged into {branch_name}"
                    )


def _clone_validation_repo(url):
    """
    Clones the authentication repository based on the url specified using the
    mirrors parameter. The repository is cloned as a bare repository
    to a the temp directory and will be deleted one the update is done.

    If repository_name isn't provided (default value), extract it from info.json.
    """
    temp_dir = tempfile.mkdtemp()
    path = Path(temp_dir, "auth_repo").absolute()
    validation_auth_repo = AuthenticationRepository(
        path=path, urls=[url], alias="Validation repository"
    )
    validation_auth_repo.clone(bare=True)
    validation_auth_repo.fetch(fetch_all=True)

    settings.validation_repo_path = validation_auth_repo.path

    validation_auth_repo.cleanup()
    return validation_auth_repo


def _get_repository_name_raise_error_if_not_defined(validation_repo, commit):
    try:
        return _get_repository_name_from_info_json(validation_repo, commit)
    except MissingInfoJsonError as e:
        raise UpdateFailedError(str(e))


def _get_repository_name_from_info_json(auth_repo, commit_sha):
    try:
        info = auth_repo.get_json(commit_sha, INFO_JSON_PATH)
        return f'{info["namespace"]}/{info["name"]}'
    except Exception:
        raise MissingInfoJsonError(
            "Error during info.json parse. If the authentication repository's path is not specified, info.json metadata is expected to be in targets/protected"
        )


def _is_unauthenticated_allowed(repository):
    return repository.custom.get("allow-unauthenticated-commits", False)


@log_on_start(
    INFO,
    "Running TUF validation of the authentication repository...",
    logger=taf_logger,
)
def _run_tuf_updater(git_updater, auth_repo_name):
    def _init_updater():
        try:
            return Updater(
                git_updater.metadata_dir,
                "metadata/",
                git_updater.targets_dir,
                "targets/",
                fetcher=git_updater,
            )
        except Exception as e:
            taf_logger.error(f"Failed to instantiate TUF Updater due to error: {e}")
            raise e

    def _update_tuf_current_revision():
        current_commit = git_updater.current_commit
        try:
            updater.refresh()
            taf_logger.debug("Validated metadata files at revision {}", current_commit)
            # using refresh, we have updated all main roles
            # we still need to update the delegated roles (if there are any)
            # and validate any target files
            current_targets = git_updater.get_current_targets()
            for target_path in current_targets:
                target_filepath = target_path.replace("\\", "/")

                targetinfo = updater.get_targetinfo(target_filepath)
                target_data = git_updater.get_current_target_data(
                    target_filepath, raw=True
                )
                targetinfo.verify_length_and_hashes(target_data)

                taf_logger.debug(
                    "Successfully validated target file {} at {}",
                    target_filepath,
                    current_commit,
                )
            return current_commit
        except Exception as e:
            metadata_expired = EXPIRED_METADATA_ERROR in type(
                e
            ).__name__ or EXPIRED_METADATA_ERROR in str(e)
            if not metadata_expired or settings.strict:
                taf_logger.error(
                    "Validation of authentication repository {} failed at revision {} due to error: {}",
                    auth_repo_name or "",
                    current_commit,
                    e,
                )
                raise UpdateFailedError(
                    f"Validation of authentication repository {auth_repo_name or ''}"
                    f" failed at revision {current_commit} due to error: {e}"
                )
            taf_logger.warning(
                f"WARNING: Could not validate authentication repository at revision {current_commit} due to error: {e}"
            )

    last_validated_commit = None
    try:
        while not git_updater.update_done():
            updater = _init_updater()
            current_commit = _update_tuf_current_revision()
            if current_commit is not None:
                last_validated_commit = current_commit
    except UpdateFailedError as e:
        return last_validated_commit, e

    return last_validated_commit, None


def _find_next_value(value, values_list):
    """
    Find the next value in the list after the given value.

    Parameters:
    - value: The value to look for.
    - values_list: The list of values.

    Returns:
    - The next value in the list after the given value, or None if there isn't one.
    """
    try:
        index = values_list.index(value)
        if index < len(values_list) - 1:  # check if there are remaining values
            return values_list[index + 1]
    except ValueError:
        pass  # value not in list
    return None


def _merge_commit(
    repository, branch, commit_to_merge, checkout=True, force_revert=False
):
    """Merge the specified commit into the given branch and check out the branch.
    If the repository cannot contain unauthenticated commits, check out the merged commit.
    """
    taf_logger.info(
        "{} Merging commit {} into branch {}",
        repository.name,
        format_commit(commit_to_merge),
        branch,
    )
    try:
        repository.checkout_branch(branch, raise_anyway=True)
    except GitError as e:
        # two scenarios:
        # current git repository is in an inconsistent state:
        # - .git/index.lock exists (git partial update got applied)
        # should get addressed in https://github.com/openlawlibrary/taf/issues/210
        # current git repository has uncommitted changes:
        # we do not want taf to lose any repo data, so we do not reset the repository.
        # for now, raise an update error and let the user manually reset the repository
        taf_logger.error(
            "Could not checkout branch {} during commit merge. Error {}", branch, e
        )
        raise UpdateFailedError(
            f"Repository {repository.name} should contain only committed changes. \n"
            f"Please update the repository at {repository.path} manually and try again."
        )

    commit_merged = False
    if force_revert:
        # check if repository already contains this commit that needs to be merged
        # and commits following it
        commits_since_last_validated = repository.all_commits_since_commit(
            commit_to_merge
        )
        if len(commits_since_last_validated):
            repository.reset_to_commit(commit_to_merge, hard=True)
            commit_merged = True

    if not commit_merged:
        repository.merge_commit(commit_to_merge)
    if checkout:
        taf_logger.info("{}: checking out branch {}", repository.name, branch)
        repository.checkout_branch(branch)



from logging import ERROR

from typing import Dict, Tuple, Any
from attr import define, field
from logdecorator import log_on_error
from taf.git import GitRepository
from taf.updater.types.update import OperationType, UpdateType
from taf.updater.updater_pipeline import (
    AuthenticationRepositoryUpdatePipeline,
    _merge_commit,
)

from pathlib import Path
from taf.log import taf_logger, disable_tuf_console_logging
import taf.repositoriesdb as repositoriesdb
from taf.utils import is_non_empty_directory, timed_run
import taf.settings as settings
from taf.exceptions import (
    ScriptExecutionError,
    UpdateFailedError,
    ValidationFailedError,
)

from taf.updater.lifecycle_handlers import (
    handle_repo_event,
    handle_update_event,
    Event,
)
from cattr import unstructure


disable_tuf_console_logging()


def _check_update_status(repos_update_data: Dict[str, Any]) -> Tuple[Event, str]:
    # helper function to set update status of update handler based on repo status.
    # if repo handlers event status changed,
    # change the update handler status
    update_status = Event.UNCHANGED
    errors = ""

    for auth_repo_name in repos_update_data:
        repo_update_data = repos_update_data[auth_repo_name]
        repo_update_status = repo_update_data["update_status"]
        if update_status != repo_update_status and update_status != Event.FAILED:
            # if one failed, then failed
            # else if one changed, then changed
            # else unchanged
            update_status = repo_update_status
        repo_error = repo_update_data["error"]
        if repo_error is not None:
            errors += str(repo_error)

    return update_status, errors


def _execute_repo_handlers(
    update_status,
    auth_repo,
    scripts_root_dir,
    commits_data,
    error,
    targets_data,
    transient_data,
):
    try:
        transient = handle_repo_event(
            update_status,
            None,
            auth_repo.library_dir,
            scripts_root_dir,
            auth_repo,
            commits_data,
            error,
            targets_data,
        )
        if transient_data is not None:
            transient_data.update(transient)
    except ScriptExecutionError as e:
        if settings.development_mode:
            _reset_to_commits_before_pull(auth_repo, commits_data, targets_data)
            error = e


def _reset_to_commits_before_pull(auth_repo, commits_data, targets_data):
    taf_logger.info(
        "In development mode. Resetting repositories to commits before pull after handler failure"
    )

    def _reset_repository(repo, commits_data):
        before_pull = commits_data["before_pull"]
        if isinstance(before_pull, dict):
            before_pull = before_pull["commit"]
        after_pull = commits_data["after_pull"]
        if isinstance(after_pull, dict):
            after_pull = after_pull["commit"]
        if before_pull == after_pull:
            return
        repo.reset_to_commit(before_pull, hard=True)

    auth_repo.checkout_branch(auth_repo.default_branch)
    _reset_repository(auth_repo, commits_data)
    auth_repo.set_last_validated_commit(commits_data["before_pull"])

    for repo_name, repo_data in targets_data.items():
        repo = repositoriesdb.get_repository(auth_repo, repo_name)
        for branch, branch_data in repo_data["commits"].items():
            repo.checkout_branch(branch)
            _reset_repository(repo, branch_data)


@define
class RepositoryConfig:
    operation: OperationType = field(converter=OperationType)
    url: str = field(
        metadata={"docs": "URL of the remote authentication repository"}, default=None
    )
    path: Path = field(
        default=None,
        converter=lambda p: Path(p).resolve() if p else None,
        metadata={"docs": "Client's authentication repository's full path"},
    )
    library_dir: Path = field(
        default=None,
        metadata={
            "docs": "Directory where client's target repositories are located. Optional."
        },
    )
    update_from_filesystem: bool = field(
        default=False,
        metadata={
            "docs": "A flag indicating if the URL is actually a file system path. Optional."
        },
    )
    expected_repo_type: UpdateType = field(
        default=UpdateType.EITHER,
        metadata={
            "docs": "Indicates if the repository is a test, official, or any type. Optional."
        },
    )
    target_repo_classes: object = field(
        default=None,
        metadata={
            "docs": "A class or dictionary used for instantiating target repositories. Optional."
        },
    )
    target_factory: object = field(
        default=None,
        metadata={
            "docs": "A git repositories factory used for instantiating target repositories. Optional."
        },
    )
    only_validate: bool = field(
        default=False,
        metadata={
            "docs": "Specifies if repositories should only be validated without being updated. Optional."
        },
    )
    validate_from_commit: str = field(
        default=None,
        metadata={"docs": "Commit from which validation should start. Optional."},
    )
    out_of_band_authentication: str = field(
        default=None,
        metadata={"docs": "Out-of-band authentication commit's SHA. Optional."},
    )
    scripts_root_dir: Path = field(
        default=None,
        metadata={
            "docs": "Local directory for script testing, not in the authentication repository. Optional."
        },
    )
    checkout: bool = field(
        default=True,
        metadata={
            "docs": "Whether to checkout last validated commits after update. Optional."
        },
    )
    excluded_target_globs: list = field(
        default=None,
        metadata={
            "docs": "Globs specifying target repositories to exclude from validation and update. Optional."
        },
    )
    strict: bool = field(
        default=False,
        metadata={"docs": "Whether update fails if a warning is raised. Optional."},
    )

    def __attrs_post_init__(self):
        if self.operation == OperationType.CLONE:
            if self.library_dir is None:
                if self.path:
                    self.library_dir = self.path.parent.parent
                else:
                    self.library_dir = Path(".").resolve()

        if self.operation == OperationType.UPDATE:
            if self.path is None:
                self.path = Path(".").resolve()
            if self.library_dir is None:
                self.library_dir = self.path.parent.parent


@log_on_error(
    ERROR,
    "{e}",
    logger=taf_logger,
    on_exceptions=UpdateFailedError,
    reraise=True,
)
@timed_run("Cloning repositories")
def clone_repository(config: RepositoryConfig):
    """
    Validate and clone an authentication repository and its target repositories, as well
    as its dependencies (linked authentication repositories and their targets) recursively.
    Arguments:
        config: RepositoryConfig instance containing all configurations.

    Side Effects:
        If only_validate is not set to True, updates authentication repository (pulls new changes) and its target
        repositories and dependencies

    Returns:
        None
    """
    settings.strict = config.strict

    if config.url is None:
        raise UpdateFailedError("URL has to be specified when cloning repositories")

    if config.path and is_non_empty_directory(config.path):
        raise UpdateFailedError(
            f"Destination path {config.path} already exists and is not an empty directory. Run `taf repo update` to update it."
        )

    config.operation = OperationType.CLONE
    return _update_or_clone_repository(config)


@log_on_error(
    ERROR,
    "{e}",
    logger=taf_logger,
    on_exceptions=UpdateFailedError,
    reraise=True,
)
@timed_run("Updating repository")
def update_repository(config: RepositoryConfig):
    """
    Validate and update an authentication repository and its target repositories, as well
    as its dependencies (linked authentication repositories and their targets) recursively.

    Arguments:
        config: RepositoryConfig instance containing all configurations.

    Side Effects:
        If only_validate is not set to True, updates authentication repository (pulls new changes) and its target
        repositories and dependencies

    Returns:
        None
    """
    settings.strict = config.strict

    # if path is not specified, name should be read from info.json
    # which is available after the remote repository is cloned and validated

    auth_repo = GitRepository(path=config.path)
    if not config.path.is_dir() or not auth_repo.is_git_repository:
        raise UpdateFailedError(
            f"{config.path} is not a Git repository. Run `taf repo clone` instead"
        )

    taf_logger.info(f"Updating repository {auth_repo.name}")

    if config.url is None:
        config.url = auth_repo.get_remote_url()
        if config.url is None:
            raise UpdateFailedError("URL cannot be determined. Please specify it")

    return _update_or_clone_repository(config)


def _update_or_clone_repository(config: RepositoryConfig):
    repos_update_data: Dict = {}
    transient_data: Dict = {}
    root_error = None
    auth_repo_name = None
    try:

        auth_repo_name, error = _update_named_repository(
            config.operation,
            config.url,
            config.path,
            config.library_dir,
            config.update_from_filesystem,
            config.expected_repo_type,
            config.target_repo_classes,
            config.target_factory,
            config.only_validate,
            config.validate_from_commit,
            None,
            repos_update_data=repos_update_data,
            transient_data=transient_data,
            out_of_band_authentication=config.out_of_band_authentication,
            scripts_root_dir=config.scripts_root_dir,
            checkout=config.checkout,
            excluded_target_globs=config.excluded_target_globs,
        )
        if error:
            raise error
    except Exception as e:
        root_error = UpdateFailedError(
            f"Update of {auth_repo_name or 'repository'} failed due to error: {e}"
        )

    update_data = {}
    if not config.excluded_target_globs:
        # after all repositories have been updated
        # update information is in repos_update_data
        if auth_repo_name is None or auth_repo_name not in repos_update_data:
            # this must mean that an error occurred
            if root_error is not None:
                raise root_error
            else:
                raise UpdateFailedError(f"Update of {auth_repo_name} failed")
        root_auth_repo = repos_update_data[auth_repo_name]["auth_repo"]

        update_status, errors = _check_update_status(repos_update_data)
        update_transient_data = _update_transient_data(
            transient_data, repos_update_data
        )

        update_data = handle_update_event(
            update_status,
            update_transient_data,
            root_auth_repo.library_dir,
            config.scripts_root_dir,
            repos_update_data,
            errors,
            root_auth_repo,
        )

    if root_error:
        raise root_error
    return unstructure(update_data)


def _update_named_repository(
    operation,
    url,
    auth_path,
    library_dir,
    update_from_filesystem,
    expected_repo_type,
    target_repo_classes=None,
    target_factory=None,
    only_validate=False,
    validate_from_commit=None,
    conf_directory_root=None,
    visited=None,
    repos_update_data=None,
    transient_data=None,
    out_of_band_authentication=None,
    scripts_root_dir=None,
    checkout=True,
    excluded_target_globs=None,
):
    """
    Arguments:
        url: URL of the remote authentication repository
        clients_library_dir: Directory where client's target repositories are located.
        auth_repo_name: Authentication repository's name
        update_from_filesystem (optional): A flag which indicates if the URL is actually a file system path
        expected_repo_type (optional): Indicates if the authentication repository which needs to be updated is
            a test repository, official repository, or if the type is not important
            and should not be validated
        target_repo_classes (optional): A class or a dictionary used when instantiating target repositories.
            See repositoriesdb load_repositories for more details
        target_factory: A git repositories factory used when instantiating target repositories.
            See repositoriesdb load_repositories for more details
        only_validate (optional): a flag that specifies if the repositories should only be validated without
            being updated
        validate_from_commit (optional): commit from which the validation should start, allowing shorter
            execution time
        visited (optional): Authentication repositories which were already processed
        repos_update_data (optional): update status, commits data, targets data of the repository which was updated
        out_of_band_authentication (optional): out-of-band authentication commit's sha
        scripts_root_dir (optional): local directory which does not have to be contained by the authentication
            repository. Used for testing purposes while developing the scripts
        checkout (optional): Whether to checkout last validated commits after update is done
        excluded_target_globs (options): globs specifying target repositories which should not get validated and updated.
        strict (optional): Whether or not update fails if a warning is raised

    The general idea of the updater is the following:
    - We have a git repository which contains the metadata files. These metadata files
    are in the 'metadata' directory
    - Clients have a clone of that repository on their local machine and want to update it
    - We don't want to simply pull the updates. We want to verify that the new commits
    (committed after the most recent one in the client's local repository)
    - For each of the new commits, we want to check if all metadata is valid. The set of
    metadata should be valid as a whole at that revision. Not only do we want to make sure
    that a metadata which is supposed to be changed was indeed updated and is valid, but
    also to make sure that if a metadata file should not be updated, it remained the same.
    - We also want to make sure that all targets metadata is valid (including the delegated roles)
    - We do not want to simply update the metadata to the latest version, without skipping
    these checks. We want to check each commit, not just the last one.
    - If we are checking a commit which is not the latest one, we do not want to report an error
    if the metadata expired. We want to make sure that that was valid at the time when the
    metadata was committed.
    - We can rely on the TUF's way of handling metadata, by using the current and previous
    directories. We just want to automatically create and update them. They should not
    remain on the client's machine.
    - We do not want to modify TUF's updater to much, but still need to get around the fact
    that TUF skips mirrors which do not have valid and/or current metadata files. Also, we
    do not simply want to find the latest metadata, we want to validate everything in-between.
    That is why the idea is to call refresh multiple times, until the last commit is reached.
    The 'GitUpdater' updater is designed in such a way that for each new call it
    loads data from a most recent commit.
    """
    if visited is None:
        visited = []
    # if there is a recursive dependency
    if url in visited:
        return
    visited.append(url)
    # at the moment, we assume that the initial commit is valid and that it contains at least root.json
    (
        update_status,
        auth_repo,
        auth_repo_name,
        commits_data,
        error,
        targets_data,
    ) = _update_current_repository(
        operation,
        url,
        auth_path,
        library_dir,
        update_from_filesystem,
        expected_repo_type,
        target_repo_classes,
        target_factory,
        only_validate,
        validate_from_commit,
        conf_directory_root,
        out_of_band_authentication,
        checkout,
        excluded_target_globs,
    )

    # if auth_repo doesn't exist, means that either clients-auth-path isn't provided,
    # or info.json is missing from protected
    if auth_repo is None:
        raise error

    # if commits_data is empty, do not attempt to load the dependencies
    # that can happen in the repository didn't exists, but could not be validated
    # and was therefore deleted
    # or if the last validated commit is not equal to the top commit, meaning that
    # the repository was updated without using the updater
    # this second case could be reworked to return the state as of the last validated commit
    # but treat the repository as invalid for now
    commits = []
    if commits_data["after_pull"] is not None:
        if commits_data["before_pull"] is not None:
            commits = [commits_data["before_pull"]]
        commits.extend(commits_data["new"])
        # TODO
        # need to handle wrong definitions and make sure that the update doesn't fail
        # for now, just take the newest commit and do not worry about updated definitions
        # latest_commit = commits[-1::]
        repositoriesdb.load_dependencies(
            auth_repo,
            library_dir=library_dir,
            commits=commits,
        )

        if update_status != Event.FAILED:
            errors = []
            # load the repositories from dependencies.json and update these repositories
            child_auth_repos = repositoriesdb.get_deduplicated_auth_repositories(
                auth_repo, commits
            ).values()

            for child_auth_repo in child_auth_repos:
                try:
                    _, error = _update_named_repository(
                        operation=OperationType.CLONE_OR_UPDATE,
                        url=child_auth_repo.urls[0],
                        auth_path=child_auth_repo.path,
                        library_dir=library_dir,
                        update_from_filesystem=update_from_filesystem,
                        expected_repo_type=expected_repo_type,
                        target_repo_classes=target_repo_classes,
                        target_factory=target_factory,
                        only_validate=only_validate,
                        validate_from_commit=validate_from_commit,
                        conf_directory_root=conf_directory_root,
                        visited=visited,
                        repos_update_data=repos_update_data,
                        transient_data=transient_data,
                        out_of_band_authentication=child_auth_repo.out_of_band_authentication,
                        scripts_root_dir=scripts_root_dir,
                        checkout=checkout,
                    )
                    if error:
                        raise error
                except Exception as e:
                    errors.append(str(e))

            if len(errors):
                errors = "\n".join(errors)
                taf_logger.error(
                    "Update of {} failed. One or more referenced authentication repositories could not be validated:\n {}",
                    auth_repo.name,
                    errors,
                )
                error = UpdateFailedError(
                    f"Update of {auth_repo.name} failed. One or more referenced authentication repositories could not be validated:\n {errors}"
                )
                update_status = Event.FAILED

        if (
            not only_validate
            and len(commits)
            and (update_status == Event.CHANGED or update_status == Event.PARTIAL)
        ):
            # when performing breadth-first update, validation might fail at some point
            # but we want to update all repository up to it
            # so set last validated commit to this last valid commit
            last_commit = commits[-1]
            # if there were no errors, merge the last validated authentication repository commit
            _merge_commit(
                auth_repo, auth_repo.default_branch, last_commit, checkout, True
            )
            # update the last validated commit
            if not excluded_target_globs:
                auth_repo.set_last_validated_commit(last_commit)

        # do not call the handlers if only validating the repositories
        # if a handler fails and we are in the development mode, revert the update
        # so that it's easy to try again after fixing the handler
        if not only_validate and not excluded_target_globs:
            _execute_repo_handlers(
                update_status,
                auth_repo,
                scripts_root_dir,
                commits_data,
                error,
                targets_data,
                transient_data,
            )
    if repos_update_data is not None:
        repos_update_data[auth_repo.name] = {
            "auth_repo": auth_repo,
            "update_status": update_status,
            "commits_data": commits_data,
            "error": error,
            "targets_data": targets_data,
        }

    repositoriesdb.clear_repositories_db()

    return auth_repo_name, error


def _update_current_repository(
    operation,
    url,
    auth_path,
    library_dir,
    update_from_filesystem,
    expected_repo_type,
    target_repo_classes,
    target_factory,
    only_validate,
    validate_from_commit,
    conf_directory_root,
    out_of_band_authentication,
    checkout,
    excluded_target_globs,
):
    updater_pipeline = AuthenticationRepositoryUpdatePipeline(
        operation,
        url,
        auth_path,
        library_dir,
        update_from_filesystem,
        expected_repo_type,
        target_repo_classes,
        target_factory,
        only_validate,
        validate_from_commit,
        conf_directory_root,
        out_of_band_authentication,
        checkout,
        excluded_target_globs,
    )
    updater_pipeline.run()
    output = updater_pipeline.output
    return (
        output.event,
        output.users_auth_repo,
        output.auth_repo_name,
        output.commits_data,
        output.error,
        output.targets_data,
    )


def _update_transient_data(
    transient_data, repos_update_data: Dict[str, str]
) -> Dict[str, Any]:
    update_transient_data = {}
    for auth_repo_name in repos_update_data:
        if auth_repo_name in transient_data:
            update_transient_data[auth_repo_name] = transient_data[auth_repo_name]
    return update_transient_data


@timed_run("Validating repository")
def validate_repository(
    auth_path,
    library_dir=None,
    validate_from_commit=None,
    excluded_target_globs=None,
    strict=False,
):
    settings.strict = strict

    auth_path = Path(auth_path).resolve()

    if library_dir is None:
        library_dir = auth_path.parent.parent
    else:
        library_dir = Path(library_dir).resolve()

    expected_repo_type = (
        UpdateType.TEST
        if (auth_path / "targets" / "test-auth-repo").exists()
        else UpdateType.OFFICIAL
    )
    settings.overwrite_last_validated_commit = True
    settings.last_validated_commit = validate_from_commit
    try:

        auth_repo_name, error = _update_named_repository(
            operation=OperationType.UPDATE,
            url=str(auth_path),
            auth_path=str(auth_path),
            library_dir=library_dir,
            update_from_filesystem=True,
            expected_repo_type=expected_repo_type,
            only_validate=True,
            validate_from_commit=validate_from_commit,
            excluded_target_globs=excluded_target_globs,
        )
        if error:
            raise error
    except Exception as e:
        raise ValidationFailedError(
            f"Validation or repository {auth_repo_name} failed due to error: {e}"
        )
    settings.overwrite_last_validated_commit = False
    settings.last_validated_commit = None
